{***************************************************}
{                                                   }
{              Made by Virus (Hungary)              }
{                                                   }
{               MailSlot Component                  }
{                                                   }
{***************************************************}

unit MailSlot;

interface

{$R *.RES}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TMailSender = record
     ComputerName: string;
     AppHandle   : THandle;
  end;

  TMailID = record
    ID_1 : integer;
    ID_2 : integer;
  end;

  TMail = record
    Sender: TMailSender;
    MailID: TMailID;
    Mail: string;
  end;

  TMailEvent    = procedure(Sender: TMailSender; Mail: string) of object;
  TSendEvent    = procedure(Successful: boolean) of object;

  EMailException = class(Exception);

  { TMailSlot}

  TMailSlot = class(TComponent)
  private
    FActive        : boolean;
    FInterval      : Cardinal;
    FMailChannel   : string;
    FMultipleMail  : boolean;
    FOnMailIncoming: TMailEvent;
    FOnMailSended  : TSendEvent;
    FWindowHandle  : HWND;
    procedure UpdateTimer;
    procedure SetMailChannel(Value: string);
    procedure SetMultipleMail(Value: boolean);
    procedure SetActive(Value: boolean);
    procedure SetInterval(Value: Cardinal);
    procedure WndProc(var Msg: TMessage);
  protected
    MailHandle : THandle;
    LastMailID : TMailID;
    function ReadMail: string;
    procedure ExtractMail(Mail: string);
    procedure CreateConnection;
    procedure CloseConnection;
    procedure Execute; dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure WriteMail(TargetComputer: string; Mail: string);
    property Active: boolean read FActive write SetActive default False;    
  published
    property Interval: Cardinal read FInterval write SetInterval default 1000;
    property MailChannel: string read FMailChannel write SetMailChannel;
    property MultipleMail: boolean read FMultipleMail write SetMultipleMail;
    property OnMailIncoming: TMailEvent read FOnMailIncoming write FOnMailIncoming;
    property OnMailSended: TSendEvent read FOnMailSended write FOnMailSended;
  end;


procedure Register;

implementation

const

{ Errors }

  ERANDOMERROR  = 'Error while try to create a custom identifier';
  ECONNECTERROR = 'Error while try to open mailslot channel';
  ECHANNELERROR = 'The "MailChannel" is not a valid value';
  ECLOSEERROR   = 'Error while try to close the mailslot';
  EWRITEERROR   = 'Error while try to send the mail';
  EPACKTOMAIL   = 'Error when the package convert to mail';
  EGETCOMPUTER  = 'Error while try to read the computer name';
  ETIMERERROR   = 'Not enough timers available';

{ Special functions }

function GetComputerNameLocal: string;
var
  Buffer : PChar;
  Size   : DWORD;
begin
  Result := '';
  Buffer := StrAlloc(MAX_COMPUTERNAME_LENGTH + 1);
  Size   := MAX_COMPUTERNAME_LENGTH + 1;
  if GetComputerName(Buffer, Size) then
    Result := String(Buffer)
  else
    raise EMailException.Create(EGETCOMPUTER);
end;

function CreateMailID: TMailID;
begin
  try
    Randomize;
    Result.ID_1 := Random(MaxInt);
    Result.ID_2 := Random(MaxInt);
  except
    raise EMailException.Create(ERANDOMERROR);
  end;
end;

function MailToPack(Mail: string): TMail;
var
  S: TStringList;
begin
  S := TStringList.Create;
  S.Text := Mail;
  Result.MailID.ID_1         := StrToInt(S.Values['ID_1']);
  Result.MailID.ID_2         := StrToInt(S.Values['ID_2']);
  Result.Sender.ComputerName := S.Values['COMPUTER'];
  Result.Mail                := S.Values['TEXT'];
end;

function PackToMail(Mail: string): string;
var
  S: TStringList;
begin
  Result := '';
  S := TStringList.Create;
  try
    try
      with S do
        begin
          Add('[IDENTIFY]');
          Add('ID_1=' + IntToStr(CreateMailID.ID_1));
          Add('ID_2=' + IntToStr(CreateMailID.ID_2));
          Add('[SENDER]');
          Add('COMPUTER=' + GetComputerNameLocal);
          Add('[MAIL]');
          Add('TEXT=' + Mail);
          Add('[END]');
          Result := s.Text;
        end;
    except
      raise EMailException.Create(EPACKTOMAIL);
    end;
  finally
    S.Free;
  end;
end;

{ TMailSlot}

function TMailSlot.ReadMail: string;
var
  Size   : DWORD;
  Buffer : PChar;
  Getted : Boolean;
  Count  : DWORD;
begin
  Result := '';
  Getted := GetMailSlotInfo(MailHandle, nil, Size, @Count, nil);
  if (Getted) and (Size <> MAILSLOT_NO_MESSAGE) and (Count > 0) then
    begin
      Buffer := StrAlloc(Size + 1);
      ReadFile(MailHandle, Buffer^, Size, Size ,nil);
      Result := String(Buffer);
      StrDispose(Buffer);
    end;
end;

procedure TMailSlot.Execute;
var
  Mail : string;
begin
  Mail := ReadMail;
  if Mail <> '' then
    begin
      ExtractMail(Mail);
      Mail := '';
    end;
end;

procedure TMailSlot.ExtractMail(Mail: string);
var
  Pack: TMail;
begin
  Pack := MailToPack(Mail);
  if (FMultipleMail) and (Pack.MailID.ID_1 = LastMailID.ID_1) and (Pack.MailID.ID_2 = LastMailID.ID_2) then
    Exit
  else
    begin
      if Assigned(FOnMailIncoming) then
        FOnMailIncoming(Pack.Sender, Pack.Mail);
      LastMailID.ID_1 := Pack.MailID.ID_1;
      LastMailID.ID_2 := Pack.MailID.ID_2;
    end;
end;

constructor TMailSlot.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FWindowHandle := AllocateHWnd(WndProc);
  FMailChannel := '';
  FMultipleMail := True;
end;

destructor TMailSlot.Destroy;
begin
  UpdateTimer;
  DeallocateHWnd(FWindowHandle);
  CloseHandle(MailHandle);
  inherited Destroy;
end;

procedure TMailSlot.CreateConnection;
var
  Handle: THandle;
begin
  try
    Handle := CreateMailSlot(PChar('\\.\mailslot\' + FMailChannel), 0, 0, nil);
    if (Handle = INVALID_HANDLE_VALUE) then
      raise EMailException.Create(ECONNECTERROR)
    else
      MailHandle := Handle;
  except
    raise EMailException.Create(ECONNECTERROR)
  end;
end;

procedure TMailSlot.CloseConnection;
begin
  try
    CloseHandle(MailHandle);
  except
    raise EMailException.Create(ECLOSEERROR);
  end;
end;

procedure TMailSlot.SetMailChannel(Value: string);
begin
  if Value <> FMailChannel then
    FMailChannel := Value;
end;

procedure TMailSlot.SetActive(Value: boolean);
begin
  if Value <> FActive then
    begin
      if Value then
        CreateConnection
      else
        CloseConnection;
      FActive := Value;
      UpdateTimer;
   end;
end;

procedure TMailSlot.SetMultipleMail(Value: boolean);
begin
  if Value <> FMultipleMail then
    FMultipleMail := Value;
end;

procedure TMailSlot.SetInterval(Value: Cardinal);
begin
  if FInterval <> Value then
    begin
      FInterval := Value;
      UpdateTimer;
    end;
end;

procedure TMailSlot.WriteMail(TargetComputer: string; Mail: string);
var
  Size             : DWORD;
  SendMailHandle   : THandle;
begin

  Mail := PackToMail(Mail);
  TargetComputer := '\\' + TargetComputer + '\mailslot\' + FMailChannel;
  SendMailHandle := CreateFile(PChar(TargetComputer),
                               GENERIC_WRITE,
                               FILE_SHARE_READ,
                               nil,
                               OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL,
                               0);

  if (SendMailHandle <> INVALID_HANDLE_VALUE) then
    begin
      Size   := Length(Mail)+1;
      WriteFile(SendMailHandle, PChar(Mail)^, Size, Size, nil);
      CloseHandle(SendMailHandle);
      if Assigned(FOnMailSended) then FOnMailSended(True);
    end
  else
    if Assigned(FOnMailSended) then FOnMailSended(False);
end;

procedure TMailSlot.UpdateTimer;
begin
  KillTimer(FWindowHandle, 1);
  if (FInterval <> 0) and FActive then
    if SetTimer(FWindowHandle, 1, FInterval, nil) = 0 then
      raise EOutOfResources.Create(ETIMERERROR);
end;

procedure TMailSlot.WndProc(var Msg: TMessage);
begin
  with Msg do
    if Msg = WM_TIMER then
      try
        Execute;
      except
        Application.HandleException(Self);
      end
    else
      Result := DefWindowProc(FWindowHandle, Msg, wParam, lParam);
end;

procedure Register;
begin
  RegisterComponents('Samples', [TMailSlot]);
end;

end.

